// funciones.cpp: Cdigo de las funciones de esta DLL
// ------------------------------------------------------------------

#include <stdafx.h>
#include "funciones.h"

// ------------------------------------------------------------------
// Funciones exportadas
// Estn reflejadas en el archivo .DEF del proyecto en el mismo orden
// ------------------------------------------------------------------

// Funciones de Histograma
int Ejecutar(int iFunc, int iNumCtr, int iNumEnt, int iNumSal, CString& strDirTemp, 
				CString* pStrDatos, CString* pStrImgEntr, CString* pStrImgSal)
{
	// Para depuracion
	#ifdef _DEBUG
	    int i;
		TRACE0("MODDLL: Modulo en ejecucion.\n");
		TRACE("Funcion: %d\nDatos: %d\nEntradas: %d\nSalidas: %d\n", iFunc, 
			iNumCtr, iNumEnt, iNumSal);
		TRACE("Dir. Temporal: %s\n", strDirTemp);
		TRACE0("Datos: ");
		for (i=0; i<iNumCtr; i++)
			TRACE("%d:%s\n", i, pStrDatos[i]);
		TRACE0("Entradas: ");
		for (i=0; i<iNumEnt; i++)
			TRACE("%d:%s\n", i, pStrImgEntr[i]);
		TRACE0("\nSalidas: ");
		for (i=0; i<iNumSal; i++)
			TRACE("%d:%s\n", i, pStrImgSal[i]);
		TRACE0("\n");
	#endif

	// Requiere una imagen de entrada, un parmetro y almenos una salida
	if ((iNumEnt!=1) || (iNumCtr!=1) || (iNumSal<1))
		return 1;

	// Lee imagen de entrada
	CString strNombre = strDirTemp + pStrImgEntr[0];
	StCabecera cab;
	BYTE* pImgEntr = LeeImagenEntrada(strNombre, &cab);
	if (pImgEntr == NULL)
		return 2;

	// Calcula histograma en valor absoulto
	long** pHisto;
	pHisto = GeneraHistograma(pImgEntr, cab.wX, cab.wY, cab.wC);
	if (pHisto == NULL)
	{
		delete pImgEntr;
		return 3;
	}

	// Obtiene valor del tipo de salida
	int iTipo;
	LPCSTR pStr = pStrDatos[0].GetBuffer(pStrDatos[0].GetLength());
	sscanf(pStr, "%d", &iTipo);
	pStrDatos[0].ReleaseBuffer();

	// Histogramas (para cada canal) de valores escalados al tipo indicado
	BYTE* pHistoE;

	// Segn el tipo de salida...
	switch (iTipo)
	{
		case 0: // Devolver cuentas en % como valores de 1 BYTE
				pHistoE = EscalaByte(pHisto, cab.wC);
				break;

		case 1: // Devolver cuentas en valor absoluto como valores WORD (2 BYTEs)
				pHistoE = EscalaWord(pHisto, cab.wC);
				break;

		case 2: // Devolver cuentas en valor absoluto como valores DWORD (4 BYTEs)
				pHistoE = EscalaDWord(pHisto, cab.wC);
				break;

		case 3: // Devolver cuentas en % como valores FLOAT (4 BYTEs)
				pHistoE = EscalaFloat(pHisto, cab.wC);
				break;

		case 4: // Devolver cuentas en valor absoluto como valores DOUBLE (8 BYTEs)
				pHistoE = EscalaDouble(pHisto, cab.wC);
				break;

		default: // Tipo no valido
				BorraHistograma(pHisto, cab.wC);
				delete pImgEntr;
				return 4;
	}

	// Hubo error al escalar
	if (pHistoE==NULL)
	{
		BorraHistograma(pHisto, cab.wC);
		delete pImgEntr;

		return 4;
	}

	// Escribir imagenes resultado
	int iRes = EscribeResultado(strDirTemp, pStrImgSal, iTipo, pHistoE, 
		iNumSal, (int)cab.wC);

	// Libera memoria
	BorraHistoEsc(pHistoE, cab.wC, iTipo);
	BorraHistograma(pHisto, cab.wC);
	delete pImgEntr;

	return iRes;
}

// ------------------------------------------------------------------
// Funciones internas
// ------------------------------------------------------------------

// Leer imagen de entrada
BYTE* LeeImagenEntrada(CString& strNombre, StCabecera* pCab)
{
	CFile file;
	BYTE* pDatos;
	DWORD dwTam;
	
	// Abrir la imagen de entrada
	if(!file.Open(strNombre, CFile::modeRead | CFile::shareDenyWrite))
		return NULL;

	// Leer cabecera de la imagen de entrada
	if (file.Read(pCab, sizeof(StCabecera)) < sizeof(StCabecera))
	{
		file.Close();
		return NULL;
	}

	// Debe tener almenos tamao 1 de ancho y de alto, y 1 canal
	if ((pCab->wX < 1) || (pCab->wY < 1) || (pCab->wC < 1) || (pCab->wB != 1))
	{
		file.Close();
		return NULL;
	}

	// Calcula tamao de la imagen y reserva memoria
	dwTam = (DWORD)(pCab->wX) * (DWORD)pCab->wY * (DWORD)pCab->wC;
	pDatos = new BYTE[dwTam];
	if (pDatos==NULL)
	{
		file.Close();
		return NULL;
	}
	
	// Leer datos de la imagen de entrada
	if (file.Read(pDatos, dwTam) < dwTam)
	{
		file.Close();
		return NULL;
	}
	
	// Cerrar archivo de entrada
	file.Close();

	return pDatos;
}

// Escribir imagenes resultado
int EscribeResultado(CString& strTemp, CString* pStrImgSal, int iTipo, 
					 BYTE* pHistoE, int iNumSal, int iNumCanales)
{	
	CFile file;
	StCabecera cab;
	BYTE *pInd=NULL;
	int iErr;
	int iSal,iCanal;
	CString strNombre;
		
	// Cabecera para las imagenes de histograma guardadas
	cab.wX = 256;
	cab.wY = 2;
	cab.wC = 1;
	
	// Segn el tipo de salida fija el tamao del pixel y crea vector
	// de ndices del tamao addecuado
	switch (iTipo)
	{
		case 0: // Devolver cuentas en %100 como valores de 1 BYTE
				{
					pInd = new BYTE[256];
					if (pInd != NULL)
						for (WORD i=0; i<256; i++)
							pInd[i] = (BYTE)i;
					cab.wB = 1;
				}
				break;
		case 1: // Devolver cuentas en valor absoluto como valores WORD (2 BYTEs)
				{
					WORD* pIndW = new WORD[256];
					if (pIndW != NULL)
						for (WORD i=0; i<256; i++)
							pIndW[i] = i;
					pInd = (BYTE*) pIndW;
					cab.wB = 2;
				}
				break;
		case 2: // Devolver cuentas en valor absoluto como valores DWORD (4 BYTEs)
				{
					DWORD* pIndDW = new DWORD[256];
					if (pIndDW != NULL)
						for (WORD i=0; i<256; i++)
							pIndDW[i] = (DWORD)i;
					pInd = (BYTE*) pIndDW;
					cab.wB = 4;
				}
				break;
		case 3: // Devolver cuentas en valor absoluto como valores FLOAT (4 BYTEs)
				{
					float* pIndF = new float[256];
					if (pIndF != NULL)
						for (WORD i=0; i<256; i++)
							pIndF[i] = (float)i;
					pInd = (BYTE*) pIndF;
					cab.wB = 4;
				}
				break;
		case 4: // Devolver cuentas en valor absoluto como valores DOUBLE (8 BYTEs)
				{
					double* pIndD = new double[256];
					if (pIndD != NULL)
						for (WORD i=0; i<256; i++)
							pIndD[i] = (double)i;
					pInd = (BYTE*) pIndD;
					cab.wB = 8;
				}
				break;
	}
			
	// Si no se pudo crear el vector de indices
	if (pInd == NULL)
		return 5;
	
	// Guarda histogramas
	iErr = 0;
	iCanal = 0;
	for (iSal=0; iSal<iNumSal; iSal++)
	{
		strNombre = strTemp + pStrImgSal[iSal];
		if (file.Open(strNombre, CFile::modeCreate | CFile::modeWrite))
		{
			TRY
			{
				// Escribir cabecera
				file.Write(&cab, sizeof(cab));
				
				// Escribir filas de indices 
				file.Write(pInd, 256*cab.wB);

				// Escribir filas de datos segn tipo de byte
				switch (iTipo)
				{
					case 0: file.Write(((BYTE**)pHistoE)[iCanal], 256*cab.wB);
							break;
					case 1: file.Write(((WORD**)pHistoE)[iCanal], 256*cab.wB);
							break;
					case 2: file.Write(((DWORD**)pHistoE)[iCanal], 256*cab.wB);
							break;
					case 3: file.Write(((float**)pHistoE)[iCanal], 256*cab.wB);
							break;
					case 4: file.Write(((double**)pHistoE)[iCanal], 256*cab.wB);
							break;
				}
				
				// Incrementa numero de canales si queda,
				// y si no, repetri el ultimo canal
				if (iCanal < iNumCanales)
					iCanal++;
			}
			CATCH( CFileException, e )
			{
				iErr = 5;
			}
			END_CATCH
			
			file.Close();

			if (iErr>0)
				break;
		}
		else break;
	}

	// Borrar vector de indices
	switch (iTipo)
	{
		case 0: delete pInd; break;
		case 1: delete (WORD*)pInd; break;
		case 2: delete (DWORD*)pInd; break;
		case 3: delete (float*)pInd; break;
		case 4: delete (double*)pInd; break;
	}

	return iErr;
}

// Generar histograma en valor absoluto
long** GeneraHistograma(BYTE* pImg, WORD wTamX, WORD wTamY, WORD wCanales)
{
	long** pHisto;
	WORD i,j,k;

	// Reservar memoria para el array de histogramas de cada canal
	pHisto = new long * [wCanales];
	if (pHisto == NULL)
		return NULL;
	
	for (i=0; i<wCanales; i++)
	{
		// Crea histograma
		pHisto[i] = new long [256];
		if (pHisto[i] != NULL)
		{
			for (j=0; j<256; j++)
				pHisto[i][j]=0L;
		}
		else
		{
			for (j=0; j<i; j++)
				delete pHisto[j];
		
			delete pHisto;
			return NULL;
		}
	}
	
	
	// Genera histogramas
	BYTE* pValor = pImg;
	for (i=0; i<wTamY; i++)
		for (j=0; j<wTamX; j++)
			for (k=0; k<wCanales; k++)
			{
				pHisto[k][*pValor]++;
				pValor++;
			}

	return pHisto;
}

// Libera memoria del histograma
void BorraHistograma(long** pHisto, WORD wCanales)
{
	for (WORD w=0; w<wCanales; w++)
		delete pHisto[w];
	
	delete pHisto;
}

// Escalar histograma a un array de bytes en %
BYTE* EscalaByte(long** pHisto, WORD wNumCanales)
{
	long lMax;
	WORD i,j;
	BYTE** pHistoB;
	double dAux;

	// Crea  array de histogramas de bytes
	pHistoB = new BYTE* [wNumCanales];
	if (pHistoB == NULL)
		return NULL;
	
	for (i=0; i<wNumCanales; i++)
	{
		// Crea histogramas de bytes
		pHistoB[i] = new BYTE [256];
		if (pHisto[i] == NULL)
		{
			for (j=0; j<i; j++)
				delete pHistoB[j];
		
			delete pHistoB;
			return NULL;
		}

		// Obtiene valor maximo para un histograma
		lMax=0L;
		for (j=0; j<256;j++)
		{
			if (lMax<pHisto[i][j])
				lMax=pHisto[i][j];
		}
		
		// Escala valores del histograma
		for (j=0; j<256;j++)
		{
			dAux = ((double)pHisto[i][j]/lMax)*100;
			pHistoB[i][j]=(BYTE)dAux;
		}
	}

	return (BYTE*)pHistoB;
}

// Escalar histograma a un array de WORDs
BYTE* EscalaWord(long** pHisto, WORD wNumCanales)
{
	WORD i,j;
	WORD** pHistoW;
	
	// Crea  array de histogramas de WORDs
	pHistoW = new WORD* [wNumCanales];
	if (pHistoW == NULL)
		return NULL;
	
	for (i=0; i<wNumCanales; i++)
	{
		// Crea histogramas de bytes
		pHistoW[i] = new WORD [256];
		if (pHistoW[i] == NULL)
		{
			for (j=0; j<i; j++)
				delete pHistoW[j];
		
			delete pHistoW;
			return NULL;
		}

		// Escala valores del histograma
		for (j=0; j<256;j++)
			pHistoW[i][j]=(WORD)pHisto[i][j];
	}

	return (BYTE*)pHistoW;
}

// Escalar histograma a un array de DWORDs
BYTE* EscalaDWord(long** pHisto, WORD wNumCanales)
{
	WORD i,j;
	DWORD** pHistoW;
	
	// Crea  array de histogramas de WORDs
	pHistoW = new DWORD* [wNumCanales];
	if (pHistoW == NULL)
		return NULL;
	
	for (i=0; i<wNumCanales; i++)
	{
		// Crea histogramas de bytes
		pHistoW[i] = new DWORD [256];
		if (pHistoW[i] == NULL)
		{
			for (j=0; j<i; j++)
				delete pHistoW[j];
		
			delete pHistoW;
			return NULL;
		}

		// Escala valores del histograma
		for (j=0; j<256;j++)
			pHistoW[i][j]=(DWORD)pHisto[i][j];
	}

	return (BYTE*)pHistoW;
}

// Escalar histograma a un array de FLOATs
BYTE* EscalaFloat(long** pHisto, WORD wNumCanales)
{
	long lMax;
	WORD i,j;
	float** pHistoF;
	double dAux;

	// Crea  array de histogramas de WORDs
	pHistoF = new float* [wNumCanales];
	if (pHistoF == NULL)
		return NULL;
	
	for (i=0; i<wNumCanales; i++)
	{
		// Crea histogramas de bytes
		pHistoF[i] = new float [256];
		if (pHistoF[i] == NULL)
		{
			for (j=0; j<i; j++)
				delete pHistoF[j];
		
			delete pHistoF;
			return NULL;
		}

		// Obtiene valor maximo para un histograma
		lMax=0L;
		for (j=0; j<256;j++)
		{
			if (lMax<pHisto[i][j])
				lMax=pHisto[i][j];
		}
		
		// Escala valores del histograma
		for (j=0; j<256;j++)
		{
			dAux = ((double)pHisto[i][j]/lMax)*100;
			pHistoF[i][j]=(float)dAux;
		}
	}

	return (BYTE*)pHistoF;
}

// Escalar histograma a un array de Doubless
BYTE* EscalaDouble(long** pHisto, WORD wNumCanales)
{
	WORD i,j;
	double** pHistoW;
	
	// Crea  array de histogramas de WORDs
	pHistoW = new double* [wNumCanales];
	if (pHistoW == NULL)
		return NULL;
	
	for (i=0; i<wNumCanales; i++)
	{
		// Crea histogramas de bytes
		pHistoW[i] = new double [256];
		if (pHistoW[i] == NULL)
		{
			for (j=0; j<i; j++)
				delete pHistoW[j];
		
			delete pHistoW;
			return NULL;
		}

		// Escala valores del histograma
		for (j=0; j<256;j++)
			pHistoW[i][j]=(double)pHisto[i][j];
	}

	return (BYTE*)pHistoW;
}

// Borrar histograma escalado segun su tipo
void BorraHistoEsc(BYTE* pHistoE, WORD wCanales, int iTipo)
{
	// Borrar vector de indices
	switch (iTipo)
	{
		case 0: // BYTE
				{
					BYTE** pHistoEB = (BYTE**)pHistoE;
					for (WORD w=0; w<wCanales; w++)
						delete pHistoEB[w];
					delete pHistoEB;
				}
				break;
		case 1: // WORD
				{
					WORD** pHistoEW = (WORD**)pHistoE;
					for (WORD w=0; w<wCanales; w++)
						delete pHistoEW[w];
					delete pHistoEW;
				}
				break;
		case 2: // DWORD
				{
					DWORD** pHistoEW = (DWORD**)pHistoE;
					for (WORD w=0; w<wCanales; w++)
						delete pHistoEW[w];
					delete pHistoEW;
				}
				break;
		case 3: // FLOAT
				{
					float** pHistoEW = (float**)pHistoE;
					for (WORD w=0; w<wCanales; w++)
						delete pHistoEW[w];
					delete pHistoEW;
				}
				break;
		case 4: // DOUBLE
				{
					double** pHistoEW = (double**)pHistoE;
					for (WORD w=0; w<wCanales; w++)
						delete pHistoEW[w];
					delete pHistoEW;
				}
				break;
	}
}