// funciones.cpp: Cdigo de las funciones de esta DLL
// ------------------------------------------------------------------

#include <stdafx.h>
#include "funciones.h"

// ------------------------------------------------------------------
// Funciones exportadas
// Estn reflejadas en el archivo .DEF del proyecto en el mismo orden
// ------------------------------------------------------------------

/*
// Funcion de ejemplo que genera una imagen constante para cada salida
int Ejecutar(int iFunc, int iNumCtr, int iNumEnt, int iNumSal, CString& strDirTemp, 
				CString* pStrDatos, CString* pStrImgEntr, CString* pStrImgSal)
{
	int	i;
	
	// Para depuracion
	#ifdef _DEBUG
		TRACE0("MODDLL: Modulo en ejecucion.\n");
		TRACE("Funcion: %d\nEntradas: %d\nSalidas: %d\n", iFunc, iNumEnt, iNumSal);
		TRACE("Dir. Temporal: %s\n", strDirTemp);
		TRACE0("Datos: ");
		for (i=0; i<iNumCtr; i++)
			TRACE("%d:%s\n", i, pStrDatos[i]);
		TRACE0("Entradas: ");
		for (i=0; i<iNumEnt; i++)
			TRACE("%d:%s\n", i, pStrImgEntr[i]);
		TRACE0("\nSalidas: ");
		for (i=0; i<iNumSal; i++)
			TRACE("%d:%s\n", i, pStrImgSal[i]);
		TRACE0("\n");
	#endif

	// Crea imagen por defecto
	stCabecera cab;
	cab.wX = IMG_X;
	cab.wY = IMG_Y;
	cab.wC = IMG_C;
	cab.wB = 1;

	BYTE dat[IMG_TAM];
	memset(dat, 0xFF, IMG_TAM);

	// Escribir imagenes resultado
	CFile file;
	BOOL bError=FALSE;
	CString strNombre;
	int iErr = 0;
	
	for (i=0; i<iNumSal; i++)
	{
		strNombre = strDirTemp + pStrImgSal[i];

		if (file.Open(strNombre, CFile::modeCreate | CFile::modeWrite))
		{
			TRY
			{
				file.Write(&cab, sizeof(cab));
				file.Write(dat, IMG_TAM);
			}
			CATCH( CFileException, e )
			{
				#ifdef _DEBUG
					TRACE("Error al escribir imagen %s", pStrImgSal[i]);
				#endif
				bError=TRUE;
			}
			END_CATCH
			
			file.Close();

			if (bError)
			{
				iErr = 2;
				break;
			}
		}
		else
		{
			iErr = 1;
			break;
		}
	}

	// Para depuracion
	#ifdef _DEBUG
		TRACE("MODDLL: Modulo acabando con cdigo %d.\n", iErr);
	#endif


	return iErr;
}
*/

// Funciones de Histograma
int Ejecutar(int iFunc, int iNumCtr, int iNumEnt, int iNumSal, CString& strDirTemp, 
				CString* pStrDatos, CString* pStrImgEntr, CString* pStrImgSal)
{
	// Requiere una imagen de entrada y almenos una salida
	if ((iNumEnt!=1) || (iNumSal<1))
		return 1;

	// Abrir la imagen de entrada
	CString strNombre = strDirTemp + pStrImgEntr[0];
	CFile file;
	if(!file.Open(strNombre, CFile::modeRead | CFile::shareDenyWrite))
		return 2;

	// Leer cabecera de la imagen de entrada
	stCabecera cab;
	if (file.Read(&cab, sizeof(cab)) < sizeof(cab))
	{
		file.Close();
		return 3;
	}

	// Debe tener almenos tamao 1 de ancho y de alto, y 1 canal
	if ((cab.wX < 1) || (cab.wY < 1) || (cab.wC < 1) || (cab.wB != 1))
	{
		file.Close();
		return 4;
	}

	// Calcula tamao de la imagen y reserva memoria
	UINT uTam = (UINT)(cab.wX * cab.wY * cab.wC);
	BYTE* pDatos = new BYTE[uTam];
	if (pDatos==NULL)
		return 5;
	
	// Leer datos de la imagen de entrada
	if (file.Read(pDatos, uTam) < uTam)
	{
		file.Close();
		return 6;
	}
	
	// Cerrar archivo de entrada
	file.Close();

	// Reservar memoria para los histogramas de cada canal en %
	int iCanales = cab.wC;
	BYTE** pHistoPCien = new BYTE * [iCanales];
	if (pHistoPCien==NULL)
	{
		delete pDatos;
		return 5;
	}

	int i,j;
	BOOL bError=FALSE;
	for (i=0; i<iCanales; i++)
	{
		pHistoPCien[i] = new BYTE[256];
		if (pHistoPCien[i]==NULL)
			bError=TRUE;
	}
	
	if (bError)
	{
		for (i=0; i<iCanales; i++)
		{
			if (pHistoPCien[i]!=NULL)
				delete pHistoPCien[i];
		}

		delete pHistoPCien;
		delete pDatos;
		return 5;
	}

	// Reservar memoria para los histogramas de cada canal
	double** pHisto = new double * [iCanales];
	if (pHisto==NULL)
	{
		for (i=0; i<iCanales; i++)
			delete pHistoPCien[i];
		delete pHistoPCien;
		delete pDatos;
		return 5;
	}

	bError=FALSE;
	for (i=0; i<iCanales; i++)
	{
		pHisto[i] = new double[256];
		if (pHisto[i]!=NULL)
		{
			for (j=0; j<256; j++)
				pHisto[i][j]=0.0;
		}
		else
			bError=TRUE;
	}
	
	if (bError)
	{
		for (i=0; i<iCanales; i++)
		{
			if (pHisto[i]!=NULL)
				delete pHisto[i];
		}

		delete pHisto;
		for (i=0; i<iCanales; i++)
			delete pHistoPCien[i];
		delete pHistoPCien;
		delete pDatos;
		return 5;
	}

	// Genera histogramas
	int k;
	int iMaxX=cab.wX;
	int iMaxY=cab.wY;
	BYTE* pValor = pDatos;
	for (j=0; j<iMaxY; j++)
	{
		for (i=0; i<iMaxX; i++)
		{
			for (k=0; k<iCanales; k++)
			{
				pHisto[k][*pValor]+=1.0;
				pValor++;
			}
		}
	}

	// Escala histogramas para obtener valores en % con relacion al valor maximo
	double dMax;
	for (i=0; i<iCanales;i++)
	{
		// Obtiene valor maximo para un histograma
		dMax=0.0;
		for (j=0; j<256;j++)
		{
			if (dMax<pHisto[i][j])
				dMax=pHisto[i][j];
		}
		
		// Escala valores del histograma
		for (j=0; j<256;j++)
			pHistoPCien[i][j]=(BYTE)((pHisto[i][j]/dMax)*100);
	}

	// Ya no hace falta la imagen
	delete pDatos;

	// Vector de indices
	BYTE pIndices[256];
	for (i=0; i<256; i++)
		pIndices[i]=i;
	
	// Cabecera para las imagenes de histograma guardadas
	cab.wX = 256;
	cab.wY = 2;
	cab.wC = 1;
	cab.wB = 1;

	// Guarda histogramas
	bError=FALSE;
	int iErr = 0;
	j=0;
	for (i=0; i<iNumSal; i++)
	{
		strNombre = strDirTemp + pStrImgSal[i];
		if (file.Open(strNombre, CFile::modeCreate | CFile::modeWrite))
		{
			TRY
			{
				// Escribir cabecera
				file.Write(&cab, sizeof(cab));
				
				// Escribir fila de indices
				file.Write(pIndices, 256);
				
				// Escribir fila de datos
				file.Write(pHistoPCien[j], 256);

				// Incrementa numero de canales si queda,
				// y si no, repetri el ultimo canal
				if (j<iCanales)
					j++;
			}
			CATCH( CFileException, e )
			{
				bError=TRUE;
			}
			END_CATCH
			
			file.Close();

			if (bError)
			{
				iErr = 8;
				break;
			}
		}
		else
		{
			iErr = 7;
			break;
		}
	}

	// Libera datos de histogramas
	for (i=0; i<iCanales; i++)
			delete pHistoPCien[i];
		delete pHistoPCien;

	for (i=0; i<iCanales; i++)
		delete pHisto[i];
	delete pHisto;
	
	return iErr;
}
// ------------------------------------------------------------------
// Funciones internas
// ------------------------------------------------------------------

